//
// Created by pcg1 on 12.01.18.
//

#include "DFO.h"
#include "GradientEnhancedModel.h"
#include "VirtualSimulator.h"
#include <iostream>
#include <iomanip>
#include <casadi/casadi.hpp>

#define FIND_POINTS1 0
#define FIND_POINTS2 1
#define INITIALIZE_MODEL 2
#define CRITICALITY_STEP_START 3
#define CRITICALITY_STEP_CHECK_CONVERGENCE 4
#define FIND_TRIAL_POINT 5
#define ACCEPTANCE_OF_TRIAL_POINT 6
#define MODEL_IMPROVEMENT_STEP_START 7
#define MODEL_IMPROVEMENT_STEP_END 8
#define TRUST_REGION_RADIUS_UPDATE_STEP 9
#define CRITICALITY_STEP_ADD_POINTS 10

VirtualSimulator vs;



Eigen::VectorXd evaluateFunctionVS(Eigen::VectorXd x, int ng) {
  Eigen::VectorXd result(1 + ng);
  result(0) = vs.evaluateFunction(x);

  Eigen::VectorXd gradients1 = vs.evaluateFunctionGradients(x);
  Eigen::VectorXd grads = gradients1.tail(ng);
  for (int i = 0; i < grads.rows(); ++i) {
    result(i + 1) = grads(i);
  }

  return result;
}




namespace Optimization {
namespace Optimizers {

DFO::DFO(Settings::Optimizer *settings,
         Optimization::Case *base_case,
         Model::Properties::VariablePropertyContainer *variables,
         Reservoir::Grid::Grid *grid,
         Logger *logger)
    : Optimizer(settings, base_case, variables, grid, logger),
      DFO_model_(
          settings->parameters().number_of_interpolation_points,
          settings->parameters().number_of_variables,
          settings->parameters().number_of_variables_with_gradients,
          base_case->GetRealVarVector(),
          settings->parameters().initial_trust_region_radius,
          settings->parameters().required_poisedness,
          settings->parameters().
          weight_model_determination_minimum_change_hessian,
          settings->parameters().weights_distance_from_optimum_lsq,
          settings) {
  // Set parameters
  if (settings->parameters().initial_trust_region_radius > 0.0)
    initial_trust_region_radius_ = settings->parameters().initial_trust_region_radius;
  else
    initial_trust_region_radius_ = 600;

  if (settings->parameters().number_of_interpolation_points > 0)
    number_of_interpolation_points_ = settings->parameters().number_of_interpolation_points;
  else
    number_of_interpolation_points_ = 21;

  if (settings->parameters().number_of_variables > 0)
    number_of_variables_ = settings->parameters().number_of_variables;
  else
    number_of_variables_ = 10;

  if (settings->parameters().required_poisedness > 0)
    required_poisedness_ = settings->parameters().required_poisedness;
  else
    required_poisedness_ = 5;

  settings_ = settings;
  varcont_ = variables;
  iterations_ = 0;
  previous_iterate_type_ = 0;
  base_case_ = new Case(base_case);
  variables_ = variables;
  base_case_->GetRealVarVector();
  base_case_->GetRealVarIdVector();
  last_action_ = -1;
  weights_distance_from_optimum_lsq_ =
      Eigen::VectorXd::Zero(settings->
          parameters().weights_distance_from_optimum_lsq.size());
  int j = 0;
  for (auto i = settings->parameters().weights_distance_from_optimum_lsq.begin();
       i != settings->parameters().weights_distance_from_optimum_lsq.end(); ++i) {
    weights_distance_from_optimum_lsq_[j] = *i;
    j++;
    if (j >= 3) {
      break;
    }
  }

  UUIDS_in_same_order_as_in_realvar_base_case_ = base_case_->GetRealVarIdVector();
  CreateScalingVector();
  Eigen::VectorXd initialStartPoint = base_case->GetRealVarVector();
  auto as = base_case->GetRealVarIdVector(); //QList<QUuid>
  auto aa = base_case->real_variables(); // QHash<QUuid, double>
  Eigen::VectorXd realvec = base_case_->GetRealVarVector();
  DFO_model_.setInitialy0(ScaleVariablesFromApplicationToAlgorithm( base_case_->GetRealVarVector()));
  vs = VirtualSimulator(settings_->parameters().test_problem_file);
  DFO_model_.initLagrangeMultipliers(vs.GetNumberOfConstraints());
  number_of_crit_step_finished_with_bad_poisedness = 0;
  number_of_crit_step_finished_without_checking_poisedness = 0;
  ng = settings_->parameters().number_of_variables_with_gradients;
  alpha = settings_->parameters().
      weight_model_determination_minimum_change_hessian;
  QList<Case *> newCases = case_handler_->RecentlyEvaluatedCases();
  number_of_tiny_improvements = 0;
  r = settings_->parameters().r;
  w = settings_->parameters().w;
  u = settings_->parameters().u;
  beta = settings_->parameters().beta;
  epsilon_c = settings_->parameters().epsilon_c;
  tau = settings_->parameters().tau;
  eta1 = settings_->parameters().eta1;
  gamma = settings_->parameters().gamma;
  gamma_inc = settings_->parameters().gamma_inc;
  trust_region_radius_inc = settings_->parameters().initial_trust_region_radius;
  trust_region_radius_max = settings_->parameters().max_trust_region_radius;
  trust_region_radius_end = settings_->parameters().end_trust_region_radius;
  notConverged = true;
  crit_steps = 0;
  accept_steps = 0;
  model_impr_steps = 0;
  refFuncVals = DFO_model_.getFvalsReference();
  refY = DFO_model_.getYReference();
  refDerivatives = DFO_model_.getDerivativeReference();
  for (int i = 0; i < number_of_interpolation_points_; ++i) {
    (*refFuncVals)(i) = -1;
  }
  number_of_new_points = 0;
  number_of_function_calls = 0;
  number_of_parallell_function_calls = 0;
  last_trial_point = Eigen::VectorXd::Zero(number_of_variables_);
  new_point = Eigen::VectorXd::Zero(number_of_variables_);
  function_evaluations = Eigen::VectorXd::Zero(number_of_interpolation_points_);
  function_evaluations.setZero();
  new_point_index = -1;
  criticality_step_iteration = 0;
  force_criticality_step = false;
  {
    Eigen::VectorXd tmp = DFO_model_.getCenterPoint();
    for (int i = 0; i < number_of_variables_; ++i) {
      last_trial_point(i) = tmp(i) + DFO_model_.GetTrustRegionRadius() * (2 * r) * 10; //far outside
    }
  }
  next_step_ = FIND_POINTS1;
  is_model_CFL = false; //false means we don't know.
  trust_region_radius_tilde = settings_->parameters().initial_trust_region_radius;;
  trust_region_radius_icb = settings_->parameters().initial_trust_region_radius;;
  rho = 0;
  multiple_new_points = false;
  number_of_points_first_set = 0;
  Eigen::VectorXd realvars = base_case_->GetRealVarVector();
  DFO_model_.findVariableMeaning(realvars , scaling_);
}
void DFO::handleEvaluatedCase(Optimization::Case *c) {
}

void DFO::iterate() {
  if (terminated){
    return;
  }

  if (iterations_ != 0){
    //Extract information from the new cases.
    if (multiple_new_points) {
      QList<Case *> cs = case_handler_->RecentlyEvaluatedCases();
      for( int i=0; i<cs.count(); ++i )
      {
        for (int j = 0; j < number_of_new_points; j++) {
          if (cs[i]->GetId() == mapIndicesToCase[j] ) {
            function_evaluations[j] = -1*(cs[i]->objective_function_value());
            break;
          }
        }
      }

    } else {
      QList<Case *> c = case_handler_->RecentlyEvaluatedCases();
      function_evaluation = -1*(c[0]->objective_function_value());
    }

    case_handler_->ClearRecentlyEvaluatedCases();
  }


  print:
  std::cout << "\033[1;34;m " << " ---------- New iterate " << iterations_ << " ---------- " << "\033[0m"
            << std::endl;
  std::cout << "\033[1;34;m " << "Fvals = \n" << "\033[0m" << *refFuncVals << "\n";
  std::cout << "\033[1;34;m " << "Y = \n" << "\033[0m" << *refY << "\n";
  if (iterations_ != 0 && iterations_ != 1 && iterations_ != 2) {
    std::cout << "\033[1;34;m " << "Best point index = \n" << "\033[0m" << DFO_model_.getBestPointIndex() << "\n";

    DFO_model_.PrintSortedBestPoint(scaling_);
  }
  std::cout << "\033[1;34;m " << "y0 = \n" << "\033[0m" << DFO_model_.getCenterPoint() << "\n";
  std::cout << "\033[1;34;m " << "Trust region radius is: " << "\033[0m" << DFO_model_.GetTrustRegionRadius()
            << std::endl;
  std::cout << "\033[1;34;m " << "Trust region radius Tilde is: " << "\033[0m" << trust_region_radius_tilde
            << std::endl;


  top:
  if (next_step_ == FIND_POINTS1) {
    new_points = DFO_model_.findFirstSetOfInterpolationPoints();
    std::cout << "Y new points:\n" << new_points <<"\n";
    std::cout << "Y:\n" << *refY << "\n";
    new_points_indicies.resize(new_points.cols());
    number_of_points_first_set = new_points.cols();

    multiple_new_points = true;
    set_next_step(FIND_POINTS2);
    goto evaluate;
  } else if (next_step_ == FIND_POINTS2) {



    // add the points.
    for (int i = 0; i < number_of_new_points; ++i) {
      //DFO_model_.SetFunctionValue(i + 1, function_evaluations[i]);
      DFO_model_.SetFunctionValueAndDerivatives(i + 1, function_evaluations(i), new_gradients.col(i));
    }

    set_next_step(INITIALIZE_MODEL);
    // find the remaining points.
    if (number_of_interpolation_points_ == number_of_function_calls) {
      // All points are found.
      number_of_new_points = 0;
      goto print;
    } else {
      number_of_new_points = number_of_interpolation_points_ - number_of_function_calls;
      new_points_indicies.resize(number_of_new_points);
      new_points.resize(number_of_variables_, number_of_new_points);
      new_points = DFO_model_.findLastSetOfInterpolationPoints();
      multiple_new_points = true;
      goto evaluate;
    }
  } else if (next_step_ == INITIALIZE_MODEL) {
    if (number_of_new_points != 0) {
      // Add the points
      for (int i = 0; i < number_of_new_points; ++i) {
        DFO_model_.SetFunctionValueAndDerivatives(
            number_of_points_first_set + i + 1,
            function_evaluations(i),
            new_gradients.col(i));
      }
    }

    DFO_model_.initializeModel();
    set_next_step(CRITICALITY_STEP_START);
    goto print;
  } else if (next_step_ == CRITICALITY_STEP_START) {
    Eigen::VectorXd gradient = DFO_model_.GetLagrangianGradient(DFO_model_.GetBestPoint());
    if ((gradient.norm() > epsilon_c && force_criticality_step == false
        && number_of_crit_step_finished_without_checking_poisedness <= 4) ) {
      DFO_model_.SetTrustRegionRadius(trust_region_radius_icb);
      set_next_step(FIND_TRIAL_POINT);
      number_of_crit_step_finished_without_checking_poisedness++;
      goto top;
    } else {
      number_of_crit_step_finished_without_checking_poisedness = 0;
      bool is_poised =
          DFO_model_.ModelImprovementAlgorithm(r * trust_region_radius_icb, new_points, new_points_indicies);
      if ((!is_poised) || (trust_region_radius_icb > u * gradient.norm())) {
        criticality_step_iteration = 0;
        if (is_poised) {
          force_criticality_step = false;
          set_next_step(CRITICALITY_STEP_CHECK_CONVERGENCE);
          goto top;
        } else {
          force_criticality_step = false;
          set_next_step(CRITICALITY_STEP_ADD_POINTS);
          multiple_new_points = true;
          goto evaluate;
        }
      } else {
        if (force_criticality_step) {
          trust_region_radius_icb = trust_region_radius_icb * gamma;
          force_criticality_step = false;
          set_next_step(CRITICALITY_STEP_START);
        } else {
          DFO_model_.SetTrustRegionRadius(trust_region_radius_icb);
          set_next_step(FIND_TRIAL_POINT);
        }
        goto top;
      }
    }
  } else if (next_step_ == CRITICALITY_STEP_ADD_POINTS) {
    Eigen::VectorXi dummyI;
    Eigen::MatrixXd dummyP;
    bool is_poised = DFO_model_.ModelImprovementAlgorithm(r * trust_region_radius_tilde, dummyP, dummyI);
    DFO_model_.update(new_points,
                      function_evaluations,
                      new_gradients,
                      new_points_indicies,
                      number_of_new_points,
                      DFO_Model::IMPROVE_POISEDNESS);
    is_poised = DFO_model_.ModelImprovementAlgorithm(r * trust_region_radius_tilde, new_points, new_points_indicies);
    set_next_step(CRITICALITY_STEP_CHECK_CONVERGENCE);
    crit_steps++;
    goto print;
  } else if (next_step_ == CRITICALITY_STEP_CHECK_CONVERGENCE) {
    Eigen::VectorXd gradient = DFO_model_.GetLagrangianGradient(DFO_model_.GetBestPoint());

    bool is_poised =
        DFO_model_.ModelImprovementAlgorithm(r * trust_region_radius_tilde, new_points, new_points_indicies);

    if (trust_region_radius_tilde <= u * gradient.norm() && is_poised) {
      // The radius have been reduced, and the gradient is now sufficient large. Proceed to find trial point.
      is_model_CFL = true;
      double temp = max(trust_region_radius_tilde, beta * gradient.norm());
      double new_trust_region_radius = min(temp, trust_region_radius_icb);
      DFO_model_.SetTrustRegionRadius(new_trust_region_radius);
      set_next_step(FIND_TRIAL_POINT);
      goto print;
    } else {
      do {
        //decrease radius. check poisedness.
        criticality_step_iteration++;
        trust_region_radius_tilde = pow(w, (criticality_step_iteration - 1)) * trust_region_radius_icb;
        is_poised =
            DFO_model_.ModelImprovementAlgorithm(r * trust_region_radius_tilde, new_points, new_points_indicies);
      } while (is_poised);

      if (trust_region_radius_tilde <= trust_region_radius_end){
        std::cout << "crit_steps: " << crit_steps << "\nmodel_impr: " << model_impr_steps << "\nacceptance: "
                  << accept_steps << "\n";
        DFO_model_.Converged(iterations_, 0, number_of_function_calls, number_of_parallell_function_calls);
        for (int i = 0; i<7;i++){
          eigen_col(new_points, DFO_model_.GetBestPoint(),i);
          new_points_indicies(i) = i+1;
        }
        multiple_new_points = true;
        terminated = true;
        goto finished;

      }

      set_next_step(CRITICALITY_STEP_ADD_POINTS);
      number_of_new_points = new_points_indicies.rows();
      multiple_new_points = true;
      goto evaluate;
    }
  } else if (next_step_ == FIND_TRIAL_POINT) {

    Eigen::VectorXd gradient = DFO_model_.GetLagrangianGradient(DFO_model_.GetBestPoint());



    if (DFO_model_.GetTrustRegionRadius() <= trust_region_radius_end) {
      std::cout << "crit_steps: " << crit_steps << "\nmodel_impr: " << model_impr_steps << "\nacceptance: "
                << accept_steps << "\n";
      DFO_model_.Converged(iterations_, 0, number_of_function_calls, number_of_parallell_function_calls);
      for (int i = 0; i<7;i++){
        eigen_col(new_points, DFO_model_.GetBestPoint(),i);
        new_points_indicies(i) = i+1;
      }
      terminated = true;
      multiple_new_points = true;
      goto finished;
    }
    new_point = DFO_model_.FindLocalOptimum();
    if ((new_point - last_trial_point).norm() <= 0.0001*DFO_model_.GetTrustRegionRadius()) {
      force_criticality_step = true;
      trust_region_radius_icb = gamma * DFO_model_.GetTrustRegionRadius();
      set_next_step(CRITICALITY_STEP_START);
      goto top;
    } else {
      last_trial_point = new_point;
      double maxDistance = DFO_model_.findLargestDistanceBetweenPointsAndOptimum();
      if (DFO_model_.norm((new_point - DFO_model_.GetBestPoint())) < tau * maxDistance) {
        // Too close.
        if (!is_model_CFL) {
          is_model_CFL = DFO_model_.isPoised(new_point, new_point_index, DFO_model_.GetTrustRegionRadius());
        }
        if (is_model_CFL) {
          set_next_step(TRUST_REGION_RADIUS_UPDATE_STEP);
          goto top;
        } else {
          set_next_step(MODEL_IMPROVEMENT_STEP_START);
          goto top;
        }
      } else {
        set_next_step(ACCEPTANCE_OF_TRIAL_POINT);
        number_of_new_points = 0;
        multiple_new_points = false;
        goto evaluate;
      }
    }
  } else if (next_step_ == ACCEPTANCE_OF_TRIAL_POINT) {
    accept_steps++;
    DFO_model_.SetTrustRegionRadiusForSubproblem(
        DFO_model_.GetTrustRegionRadius());
    int t = DFO_model_.findPointToReplaceWithNewOptimum(new_point);
    rho = (DFO_model_.GetFunctionValue(DFO_model_.getBestPointIndex()) - function_evaluation)
        / (DFO_model_.evaluateQuadraticModel(DFO_model_.GetBestPoint())
            - DFO_model_.evaluateQuadraticModel(new_point));

    if (!is_model_CFL) {
      Eigen::VectorXd dummyVec(number_of_variables_);
      dummyVec.setZero();
      int dummyInt = 0;
      is_model_CFL = DFO_model_.isPoised(dummyVec, dummyInt, DFO_model_.GetTrustRegionRadius());
    }

    if ((rho >= eta1) || (is_model_CFL && rho > 0)) {
      DFO_model_.update(new_point, function_evaluation, new_gradient, t, DFO_Model::INCLUDE_NEW_OPTIMUM);
      set_next_step(TRUST_REGION_RADIUS_UPDATE_STEP);
      goto print;
    } else {
      int index = DFO_model_.isPointAcceptable(new_point);
      if (index != -1) {
        DFO_model_.update(new_point, function_evaluation, new_gradient, index, DFO_Model::INCLUDE_NEW_POINT);
      }
      set_next_step(MODEL_IMPROVEMENT_STEP_START);
      goto top;
    }
  } else if (next_step_ == MODEL_IMPROVEMENT_STEP_START) {
    DFO_model_.modelImprovementStep(new_point, new_point_index);
    if (new_point_index == -1) {
      rho = 0;
      force_criticality_step = true;
      is_model_CFL = true;
      set_next_step(TRUST_REGION_RADIUS_UPDATE_STEP);
      goto top;
    } else {
      set_next_step(MODEL_IMPROVEMENT_STEP_END);
    }
  } else if (next_step_ == MODEL_IMPROVEMENT_STEP_END) {
    DFO_model_.update(new_point, function_evaluation, new_gradient, new_point_index, DFO_Model::INCLUDE_NEW_POINT);
    model_impr_steps++;
    set_next_step(TRUST_REGION_RADIUS_UPDATE_STEP);
    goto print;
  } else if (next_step_ == TRUST_REGION_RADIUS_UPDATE_STEP) {
    if (rho >= eta1) {
      if (function_evaluation > DFO_model_.GetFunctionValue(DFO_model_.getBestPointIndex())) {
        trust_region_radius_icb = gamma * DFO_model_.GetTrustRegionRadius();
      } else {
        double tmp = std::min(gamma_inc * DFO_model_.GetTrustRegionRadius(), trust_region_radius_max);
        double weight = 0;
        trust_region_radius_icb = weight * DFO_model_.GetTrustRegionRadius() + (1 - weight) * tmp;
        rho = 0;
      }
    } else {
      if (!is_model_CFL) {
        Eigen::VectorXd dummyVec(number_of_variables_);
        dummyVec.setZero();
        int dummyInt = 0;
        is_model_CFL = DFO_model_.isPoised(dummyVec, dummyInt, DFO_model_.GetTrustRegionRadius());
      }
      if (is_model_CFL) {
        trust_region_radius_icb = gamma * DFO_model_.GetTrustRegionRadius();
      } else {
        trust_region_radius_icb = DFO_model_.GetTrustRegionRadius();
      }
    }
    set_next_step(CRITICALITY_STEP_START);
    goto top;

  } else {
    cout << "This should never be ran. \n";
    cin.get();
  }


  evaluate:
  if (multiple_new_points) {
    number_of_new_points = new_points_indicies.rows();
    new_gradients.resize(ng, number_of_new_points);
    mapIndicesToCase.resize(number_of_new_points);
    function_evaluations.resize(number_of_new_points);
    for (int i = 0; i < number_of_new_points; ++i) {
      number_of_function_calls++;
      Optimization::Case *new_case = new Optimization::Case(base_case_);
      ConvertPointToCase(new_points.col(i) + DFO_model_.getCenterPoint(),new_case);
      mapIndicesToCase[i] =  new_case->GetId();
      case_handler_->AddNewCase(new_case);
    }
    number_of_parallell_function_calls++;
  } else {
    Optimization::Case *new_case = new Optimization::Case(base_case_);
    ConvertPointToCase(new_point + DFO_model_.getCenterPoint(),new_case);
    mapIndexToCase =  new_case->GetId();
    //add it to the handler!
    case_handler_->AddNewCase(new_case);
    number_of_function_calls++;
    number_of_parallell_function_calls++;
  }
  iterations_++;

  finished:
  std::cout << "end of iterate\n";
}

std::string DFO::get_action_name(int a) {
  switch (a) {
    case 0: return "FIND_POINTS1 							";
    case 1: return "FIND_POINTS2 							";
    case 2: return "INITIALIZE_MODEL 						";
    case 3: return "CRITICALITY_STEP_START 				    ";
    case 4: return "CRITICALITY_STEP_CHECK_CONVERGENCE    	";
    case 5: return "FIND_TRIAL_POINT 						";
    case 6: return "ACCEPTANCE_OF_TRIAL_POINT 				";
    case 7: return "MODEL_IMPROVEMENT_STEP_START 			";
    case 8: return "MODEL_IMPROVEMENT_STEP_END 		    	";
    case 9: return "TRUST_REGION_RADIUS_UPDATE_STEP 		";
    case 10: return "CRITICALITY_STEP_ADD_POINTS 			";
    default:return "Not a valid state                       ";
  }
}

void DFO::set_next_step(int a) {
  std::cout << "From " << "\033[1;" + color_from + "m " << get_action_name(next_step_) << "\033[0m";
  next_step_ = a;
  std::cout << "\tTo " << "\033[1;" + color_to + ";m " << get_action_name(next_step_) << "\033[0m" << std::endl;
}

Eigen::VectorXd DFO::ScaleVariablesFromAlgorithmToApplication(Eigen::VectorXd point) {
  Eigen::VectorXd ret(point.rows());
  for (int i = 0; i < number_of_variables_; i++){
    ret[i] = scaling_[i]*point[i];
  }
  return ret;
}
Eigen::VectorXd DFO::ScaleVariablesFromApplicationToAlgorithm(Eigen::VectorXd point) {
  Eigen::VectorXd ret(number_of_variables_);
  for (int i = 0; i < number_of_variables_; i++){
    ret[i] = point[i]/scaling_[i];
  }
  return ret;
}


void DFO::ConvertPointToCase(Eigen::VectorXd point,Optimization::Case* new_case) {
  Eigen::VectorXd scaled(number_of_variables_);
  scaled = ScaleVariablesFromAlgorithmToApplication(point);

  new_case->SetRealVarValues(scaled);
  new_case->set_objective_function_value(
      std::numeric_limits<double>::max());
}

void DFO::CreateScalingVector() {
  scaling_ = Eigen::VectorXd::Zero(number_of_variables_);
  for (int i = 0; i < number_of_variables_; ++i) {
    Model::Properties::Property::PropertyInfo
        propinfo = variables_->
        GetContinousVariable(
        UUIDS_in_same_order_as_in_realvar_base_case_[i])->propertyInfo();
    if (propinfo.coord == Model::Properties::Property::Coordinate::x) {
      scaling_[i] = 10000.0;
    } else if (propinfo.coord == Model::Properties::Property::Coordinate::y) {
      scaling_[i] = 10000.0;
    } else if (propinfo.coord == Model::Properties::Property::Coordinate::z) {
      scaling_[i] = 10000.0 / 30.0;
    } else {
      std::cout << "The variable type is not coordinate xyz... no scaling applied\n";
      std::cin.get();
    }
  }
}


}
}

