#include "DFO_Model.h"
namespace Optimization {
namespace Optimizers {

bool DFO_Model::cmp(Eigen::VectorXd a, Eigen::VectorXd b) {
  return (norm(a.topRows(a.rows() - 1))) > (norm(b.topRows(b.rows() - 1)));
}

static int NormType = 0;

double norm2(Eigen::VectorXd a) {
  if (NormType == 0) {
    return (a).lpNorm<Infinity>();
  } else if (NormType == 2) {
    return (a).norm();
  }
}

bool cmp2(Eigen::VectorXd a, Eigen::VectorXd b) {
  return (norm2(a.topRows(a.rows() - 1))) > (norm2(b.topRows(b.rows() - 1)));
}

bool DFO_Model::isApproxZero(double value, double zeroLimit) {
  if (std::abs(value) <= zeroLimit)
    return true;
  return false;
}

int DFO_Model::kroneckerDelta(int i, int j) {
  if (i == j)
    return 1;
  return 0;
}

int DFO_Model::sign(double value) {
  if (value >= 0) {
    return 1;
  }
  return -1;
}



void DFO_Model::updateQuadraticModelNew(Eigen::VectorXd yNew, double fvalNew, unsigned int t) {
  double valModelOld = evaluateQuadraticModel(yNew);
  double diff = fvalNew - valModelOld;
  double c;
  Eigen::VectorXd grad(n);
  Eigen::MatrixXd hess(n, n);
  createLagrangePolynomial(t, c, grad, hess);

  constant += diff * c;
  gradient += diff * grad;
  hessian += diff * hess;

}


DFO_Model::DFO_Model(unsigned int m,
                     unsigned int n,
                     unsigned int ng,
                     Eigen::VectorXd y0,
                     double rhoBeg,
                     double lambda,
                     double weight_objective_minimum_change,
                     QList<double> weights_derivatives,
                     Settings::Optimizer *settings)
    : subproblem(settings),
      enhancedModel(n, m, ng, weights_derivatives, weight_objective_minimum_change),
      normType(settings->parameters().norm_type) {
  NormType = settings->parameters().norm_type;
  this->m = m;
  this->n = n;
  this->ng = ng;
  Eigen::MatrixXd wtf12(10, 10);
  Eigen::MatrixXd wtf11(20, 20);
  Eigen::MatrixXd wtf(m + n + 1, m + n + 1);
  Eigen::MatrixXd wtf2;
  wtf2.resize(m + n + 1, m + n + 1);
  this->Winv.resize(m + n + 1, m + n + 1);
  this->Winv = Eigen::MatrixXd::Zero(m + n + 1, m + n + 1);
  this->W = Eigen::MatrixXd::Zero(m + n + 1, m + n + 1);
  this->y0 = Eigen::VectorXd::Zero(n);

  this->r = settings->parameters().r;
  int j = 0;

  copyOfStartingPoint = Eigen::VectorXd::Zero(n);
  for (auto i = settings->parameters().starting_point.begin(); i != settings->parameters().starting_point.end(); ++i) {
    this->y0[j] = *i;
    this->copyOfStartingPoint[j] = *i;
    j++;
    if (j >= n) {
      break;
    }
  }

  lagabsvalMin = settings->parameters().min_lagrange_abs_val;

  this->rho = rhoBeg;
  this->lambda = lambda;

  this->bestPoint = Eigen::VectorXd::Zero(n);
  this->bestPointAllTime = Eigen::VectorXd::Zero(n);
  this->bestPointAllTimeFunctionValue = std::numeric_limits<double>::max();
  this->Y = Eigen::MatrixXd::Zero(n, m);
  this->derivatives = Eigen::MatrixXd(ng, m);
  this->derivativeAtCenterpoint = Eigen::VectorXd(ng);
  this->fvals = Eigen::VectorXd(m);
  if (m >= n + 2) {
    this->Xi = Eigen::MatrixXd::Zero(n + 1, m);
    this->Upsilon = Eigen::MatrixXd::Zero(n + 1, n + 1);
    this->Z = Eigen::MatrixXd::Zero(m, m - n - 1);
    this->S = Eigen::DiagonalMatrix<double, Eigen::Dynamic>(m - n - 1);
    this->S.diagonal().setOnes();
  }
  if (m > 2 * n + 1) {
    this->sigmas = Eigen::VectorXi(n);
    this->ps = Eigen::VectorXi(m - 2 * n - 1);
    this->qs = Eigen::VectorXi(m - 2 * n - 1);
    this->is = Eigen::VectorXi(m - 2 * n - 1);
  }

  this->gradient = Eigen::VectorXd::Zero(n);
  this->centerPoint = Eigen::VectorXd::Zero(n);
  this->hessian = Eigen::MatrixXd::Zero(n, n);
  this->Gamma = Eigen::MatrixXd::Zero(n, n);
  this->gammas = Eigen::VectorXd::Zero(m);
  this->initialInterpolationPointsFound = false;
  this->modelInitialized = false;
  this->settings_ = settings;
  this->isModelCFL = -1;

}

Eigen::MatrixXd DFO_Model::findFirstSetOfInterpolationPoints() {
  int numberOfPointsFound = 0;
  if (m < n + 2) {
    for (int i = 1; i <= n; ++i) {
      if (i <= (m - 1)) {
        Y(i - 1, i) += rho;
      }
    }
    if (m >= n) {
      for (int i = 1; i < m - n; ++i) {
        Y(i - 1, i + n) -= rho;
      }
    }
    numberOfPointsFound = m;
    initialInterpolationPointsFound = true;
  } else if (m >= 2 * n + 1 && m <= (n + 1) * (n + 2) * 0.5) {
    for (int i = 1; i <= n; ++i) {
      Y(i - 1, i) += rho;
      Y(i - 1, i + n) -= rho;
    }
    numberOfPointsFound = 2 * n + 1;
  } else if (m >= n + 2 && m <= 2 * n) {
    for (int i = 1; i <= n; ++i) {
      Y(i - 1, i) += rho;
    }
    for (int i = 1; i < m - n; ++i) {
      Y(i - 1, i + n) -= rho;
    }
    numberOfPointsFound = m;
    initialInterpolationPointsFound = true;
  } else {
    std::cout << "Invalid value of m = " << m << ". Choose  n+2 <= m <= (n+1)(n+2)/2. Recommended: m = 2n+1."
              << std::endl;
    std::cin.get();
    std::exit(1);
  }
  return Y.block(0, 0, n, numberOfPointsFound);
}

Eigen::MatrixXd DFO_Model::findLastSetOfInterpolationPoints() {
  if (m > 2 * n + 1) {
    // Calculate the sigmas
    for (int i = 1; i <= n; ++i) {
      if (fvals[i + n] < fvals[i])
        sigmas[i - 1] = -1;
      else
        sigmas[i - 1] = 1;
    }

    // Find the last set of interpolation points, while also storing the ps, qs and is for later usage.
    int j = 2 * n + 2;
    int l = 1; //Number of cycles  (i.e. number of times i has become  i == 3 * n + 2)
    int p;
    int q;
    int index = 0;

    for (int i = 2 * n + 2; i <= m; ++i) {

      if (j >= 3 * n + 2) {
        j = j - n;
        l++;
      }
      p = (j - 2 * n - 1);

      if (p + l >= 1 && p + l <= n)
        q = p + l;
      else
        q = p + l - n;

      ps[index] = p;
      qs[index] = q;
      is[index] = i;

      Y(p - 1, i - 1) += rho * sigmas[p - 1];
      Y(q - 1, i - 1) += rho * sigmas[q - 1];

      index++;
      j++;

    }
  } else {
    std::cout << "findLastSetOfInterpilationPoints() was called when m <= 2*n + 1" << std::endl;
  }
  initialInterpolationPointsFound = true;
  return Y.block(0, 2 * n + 1, n, m - (2 * n + 1));
}

void DFO_Model::initializeModel() {
  derivativeAtCenterpoint = derivatives.col(0);
  Winv = calculateWExplicitly();
  createW();
  modelInitialized = true;
  bestPointIndex = 1;
  bestPoint = Y.col(0);

}

void DFO_Model::update(Eigen::MatrixXd yNews,
                       Eigen::VectorXd fvalNews,
                       Eigen::MatrixXd gradients,
                       Eigen::VectorXi indicies,
                       int numberOfPoints,
                       UpdateReason updateReason) {
  for (int i = 0; i < numberOfPoints; ++i) {
    update(yNews.col(i), fvalNews(i), gradients.col(i), indicies(i), updateReason);
  }
}

void DFO_Model::update(Eigen::VectorXd yNew,
                       double fvalNew,
                       Eigen::VectorXd grad,
                       unsigned int t,
                       UpdateReason updateReason) {
  int oldBestPointIndex = bestPointIndex;
  int oldBestFval = fvals[bestPointIndex - 1];

  if (updateReason == INCLUDE_NEW_OPTIMUM) {
    lagrangeMultipliers = subproblem.getLagrangeMultipliers();

    if (fvalNew < bestPointAllTimeFunctionValue) {
      bestPointAllTimeFunctionValue = fvalNew;
      bestPointAllTime = yNew;
    }
  }

  if (updateReason == IMPROVE_POISEDNESS) {
    if (fvalNew < bestPointAllTimeFunctionValue) {
      bestPointAllTimeFunctionValue = fvalNew;
      bestPointAllTime = yNew;
    }
  }
  if (updateReason == INCLUDE_NEW_POINT) {
    if (fvalNew < bestPointAllTimeFunctionValue) {
      bestPointAllTimeFunctionValue = fvalNew;
      bestPointAllTime = yNew;
    }
  }

  if (updateReason == FORCED_IMPROVE_MODEL) {

  } else {
    if (fvalNew < fvals[bestPointIndex - 1] && updateReason == INCLUDE_NEW_OPTIMUM) {
      bestPoint = yNew;
      bestPointIndex = t;
    }
  }

  eigen_col(Y, yNew, t - 1);
  fvals(t - 1) = fvalNew;
  if (ng > 0) {
    eigen_col(derivatives, grad, t - 1);
  }
  updateQuadraticModelNew(yNew, fvalNew, t);

  if ((updateReason == IMPROVE_POISEDNESS || updateReason == INCLUDE_NEW_POINT)) {
    if (t == oldBestPointIndex && fvalNew > oldBestFval) { // removing optimum :(
      bestPointIndex = 1;
      for (int j = 2; j <= m; ++j) {
        if (fvals[j - 1] < fvals[bestPointIndex - 1]) {
          bestPointIndex = j;
        }
      }
      bestPoint = Y.col(bestPointIndex - 1);
    }
  }
  Winv = calculateWExplicitly();
  createW();

  isModelCFL = -1;

}

double DFO_Model::evaluateQuadraticModel(Eigen::VectorXd point) {
  enhancedModel.ComputeModel(Y, derivatives, derivativeAtCenterpoint, fvals, y0, bestPoint, rho, r, ng);

  double e_c = 0;
  Eigen::VectorXd e_g(n);
  Eigen::MatrixXd e_h(n, n);
  enhancedModel.GetModel(e_c, e_g, e_h);
  double val = e_c + point.transpose() * e_g + 0.5 * point.transpose() * e_h * point;
  return val;
}

void DFO_Model::shiftCenterPointOfQuadraticModelNew(Eigen::VectorXd s) {
  double tmp1 = gradient.transpose() * s;
  double tmp2 = (0.5 * (s.transpose() * hessian) * s);
  constant += tmp1 + tmp2;
  gradient += hessian * s;
  for (int i = 1; i <= m; ++i) {
    eigen_col(Y, Y.col(i - 1) - s, i - 1);
  }
  bestPoint -= s;
  bestPointAllTime -= s;
  y0 += s;
}


void DFO_Model::findWorstPointInInterpolationSet(Eigen::VectorXd &dNew, int &indexOfWorstPoint) {
  double worstPoisedness = 0;
  Eigen::VectorXd poisedness(m);
  int index = -1;
  double c;
  Eigen::VectorXd grad(n);
  Eigen::MatrixXd hess(n, n);

  // Creating the lagrange polynomial.
  for (int t = 1; t <= m; ++t) {
    createLagrangePolynomial(t, c, grad, hess);

    // Find min and max of l_t(x)
    subproblem.setConstant(c);
    subproblem.setGradient(grad);
    subproblem.setHessian(hess);
    vector<double> xsolMax;
    vector<double> fsolMax;
    vector<double> xsolMin;
    vector<double> fsolMin;
    //PrintLagrangePolynomial(t);
    subproblem.Solve(xsolMax, fsolMax, (char *) "Maximize", y0, bestPoint, Y.col(t - 1));
    subproblem.Solve(xsolMin, fsolMin, (char *) "Minimize", y0, bestPoint, Y.col(t - 1));
    poisedness(t - 1) = std::max(abs(fsolMax[0]), abs(fsolMin[0]));

    Eigen::VectorXd d1(n);
    Eigen::VectorXd d2(n);
    for (int i = 0; i < n; ++i) {
      d1[i] = xsolMax[i];
      d2[i] = xsolMin[i];
    }

    double temp = 0;
    if ((abs(fsolMax[0]) >= abs(fsolMin[0])) && abs(fsolMax[0]) >= worstPoisedness) {
      worstPoisedness = abs(fsolMax[0]);
      for (int i = 0; i < xsolMax.size(); ++i) {
        dNew[i] = xsolMax[i];
      }
      index = t;
    } else if ((abs(fsolMin[0]) > abs(fsolMax[0])) && abs(fsolMin[0]) >= worstPoisedness) {
      worstPoisedness = abs(fsolMin[0]);
      for (int i = 0; i < xsolMin.size(); ++i) {
        dNew[i] = xsolMin[i];
      }
      index = t;
    }
  }
  if (worstPoisedness > lambda) {

    if (index == bestPointIndex) {
      int k = -1;
      double tmp = -1;
      for (int j = 1; j <= m; j++) {
        if (poisedness[j - 1] > lambda && poisedness[j - 1] > tmp && j != bestPointIndex) {
          k = j;
          tmp = poisedness[j - 1];
        }
      }
      if (k != -1) {
        indexOfWorstPoint = k;
        indexOfWorstPoint = -1;
        return;
      } else {
        indexOfWorstPoint = -1;
        return;
      }

    } else {
      indexOfWorstPoint = index;
    }

    createLagrangePolynomial(indexOfWorstPoint, c, grad, hess);
    subproblem.setConstant(c);
    subproblem.setGradient(grad);
    subproblem.setHessian(hess);
    vector<double> xsolMax;
    vector<double> fsolMax;
    vector<double> xsolMin;
    vector<double> fsolMin;

    subproblem.SetTrustRegionRadius(GetTrustRegionRadius() * 1);
    subproblem.Solve(xsolMax, fsolMax, (char *) "Maximize", y0, bestPoint, bestPoint);
    subproblem.Solve(xsolMin, fsolMin, (char *) "Minimize", y0, bestPoint, bestPoint);

    if ((abs(fsolMax[0]) >= abs(fsolMin[0]))) {
      for (int i = 0; i < xsolMax.size(); ++i) {
        dNew[i] = xsolMax[i];
      }
    } else {
      for (int i = 0; i < xsolMin.size(); ++i) {
        dNew[i] = xsolMin[i];
      }
    }
  } else {
    indexOfWorstPoint = -1; // Indicates that the required poisedness is already achieved
  }
  std::cout << "Required poisedness: " << lambda << "\nPoisedness: " << worstPoisedness << "\n";

}


void DFO_Model::calculateAMatrix(Eigen::MatrixXd &A, Eigen::MatrixXd &Ycopy) {
  int elem = 0;
  /// Constant
  for (int i = 1; i <= m; ++i) {
    A(i - 1, 0) = 1;
  }
  elem++;
  /// Linear
  for (int i = 1; i <= n; ++i) {
    if (elem < m) {
      for (int j = 1; j <= m; ++j) {
        A(j - 1, elem) = Ycopy(i - 1, j - 1);
      }
      elem++;
    } else {
      break;
    }
  }
  /// Squared
  for (int i = 1; i <= n; ++i) {
    if (elem < m) {
      for (int j = 1; j <= m; ++j) {
        A(j - 1, elem) = Ycopy(i - 1, j - 1) * Ycopy(i - 1, j - 1);
      }
      elem++;
    } else {
      break;
    }
  }
  /// Cross terms
  int rows = n;
  int it = 1;
  int col = 1;
  int t = 2;
  for (int k = 1; k <= n - 1; ++k) {
    for (int i = t; i <= n - 1; ++i) {
      for (int j = 1; j <= m; ++j) {
        if (elem < m) {
          double x1 = Ycopy(k - 1, j - 1);
          double x2 = Ycopy(i - 1, j - 1);
          A(j - 1, elem) = x1 * x2;
        } else {
          break;
        }
      }
      elem++;
    }
    t++;
  }
}



int DFO_Model::findPointToReplaceWithNewOptimum(Eigen::VectorXd yNew) {
  // Create the w vector
  Eigen::VectorXd w(n + m + 1);
  for (int i = 1; i <= m; ++i) {
    w(i - 1) = 0.5 * std::pow((Y.col(i - 1)).transpose() * (yNew), 2);
  }
  eigen_tail(w, yNew, n);
  w(m) = 1;

  Eigen::VectorXd Hw = Eigen::VectorXd::Zero(m + n + 1);
  Hw = Winv * w;

  int indexToBeReplaced = 1;
  double currentMax = -1;
  for (int i = 1; i <= m; ++i) {
    if (i == bestPointIndex) {
      continue;
    }
    double distance = norm((bestPoint - Y.col(i - 1)));
    double distanceWeight = distance;

    double lagval = std::abs((Hw)(i - 1));
    double value = distanceWeight * lagval;
    if (value >= currentMax) {
      indexToBeReplaced = i;
      currentMax = value;
    }
  }

  std::cout << "Point selected: " << indexToBeReplaced << "\n";

  return indexToBeReplaced;
}

int DFO_Model::getBestPointIndex() {
  return bestPointIndex;
}

void DFO_Model::SetFunctionValue(int t, double value) {
  fvals[t - 1] = value;
}

void DFO_Model::SetFunctionValueAndDerivatives(int t, double value, Eigen::VectorXd grad) {
  fvals(t - 1) = value;
  for (int i = 0; i < ng; ++i) {
    derivatives(i, t - 1) = grad(i);
  }
}

void DFO_Model::SetTrustRegionRadiusForSubproblem(double radius) {
  subproblem.SetTrustRegionRadius(radius);
}

Eigen::VectorXd DFO_Model::FindLocalOptimum() {
  Eigen::VectorXd localOptimum(n);
  vector<double> xsol;
  vector<double> fsol;
  /// The enhanced model;
  enhancedModel.ComputeModel(Y, derivatives, derivativeAtCenterpoint, fvals, y0, bestPoint, rho, r, ng);
  double e_c = 0;
  Eigen::VectorXd e_g(n);
  Eigen::MatrixXd e_h(n, n);
  enhancedModel.GetModel(e_c, e_g, e_h);
  subproblem.setHessian(e_h);
  subproblem.setGradient(e_g);
  subproblem.setConstant(e_c);
  subproblem.SetTrustRegionRadius(rho);
  subproblem.Solve(xsol, fsol, (char *) "Minimize", y0, bestPoint, bestPoint);
  for (int i = 0; i < n; i++) {
    localOptimum[i] = xsol[i];
  }
  return localOptimum;
}


double DFO_Model::findLargestDistanceBetweenPointsAndOptimum() {
  int t = -1;
  double maxDistance = -1;
  for (int i = 0; i < m; ++i) {
    double dist = norm(Y.col(i) - bestPoint);
    if (dist > maxDistance) {
      t = i + 1;
      maxDistance = dist;
    }
  }
  return maxDistance;
}
double DFO_Model::ComputeLagrangePolynomial(int t, Eigen::VectorXd point) {
  double c;
  Eigen::VectorXd grad(n);
  Eigen::MatrixXd hess(n, n);
  createLagrangePolynomial(t, c, grad, hess);
  double val = c + grad.transpose() * point + 0.5 * point.transpose() * hess * point;
  return val;
}

double DFO_Model::PrintLagrangePolynomial(int t) {

  double c;
  Eigen::VectorXd grad(n);
  Eigen::MatrixXd hess(n, n);
  createLagrangePolynomial(t, c, grad, hess);
  std::cout << "Lagrange polynomial ----------------------- " << t << "\n";
  std::cout << "c = " << c << std::endl;
  std::cout << "gradient = " << std::endl << grad << std::endl;
  std::cout << "hessian = " << std::endl << hess << std::endl;
}

Eigen::VectorXd DFO_Model::FindLocalOptimumOfAbsoluteLagrangePolynomial(int t) {
  double c;
  Eigen::VectorXd grad(n);
  Eigen::MatrixXd hess(n, n);
  createLagrangePolynomial(t, c, grad, hess);

  // Find min and max of l_t(x)
  subproblem.setConstant(c);
  subproblem.setGradient(grad);
  subproblem.setHessian(hess);
  vector<double> xsolMax;
  vector<double> fsolMax;
  vector<double> xsolMin;
  vector<double> fsolMin;
  subproblem.Solve(xsolMax, fsolMax, (char *) "Maximize", y0, bestPoint, Y.col(t - 1));
  subproblem.Solve(xsolMin, fsolMin, (char *) "Minimize", y0, bestPoint, Y.col(t - 1));

  Eigen::VectorXd optimum(n);
  for (int i = 0; i < xsolMax.size(); ++i) {
    if (abs(fsolMax[0]) >= abs(fsolMin[0])) {
      optimum[i] = xsolMax[i];
    } else {
      optimum[i] = xsolMin[i];
    }
  }
  return optimum;
}

Eigen::VectorXi DFO_Model::GetInterpolationPointsSortedByDistanceFromBestPoint() {
  std::vector<Eigen::VectorXd> tmp;
  for (int i = 0; i < m; ++i) {
    Eigen::VectorXd t(n + 1);
    for (int j = 0; j < n; ++j) {
      t(j) = Y(j, i) - Y(j, bestPointIndex - 1);
    }
    t(n) = i + 1;
    tmp.push_back(t);
  }
  std::sort(tmp.begin(), tmp.end(), cmp2);
  Eigen::VectorXi indicesSortedByDescendingNorm(m);
  for (int i = 0; i < m; ++i) {
    indicesSortedByDescendingNorm[i] = tmp[i][n];
  }
  return indicesSortedByDescendingNorm;
}

bool DFO_Model::FindPointToReplaceWithPointOutsideScaledTrustRegion(int t, Eigen::VectorXd &dNew) {
  subproblem.SetTrustRegionRadius(rho);
  dNew = FindLocalOptimumOfAbsoluteLagrangePolynomial(t);
  if (std::abs(ComputeLagrangePolynomial(t, dNew)) > lambda) {
    return true;
  }
  return false;
}


int DFO_Model::isPointAcceptable(Eigen::VectorXd point) {
  // Create the w vector
  Eigen::VectorXd w(n + m + 1);
  for (int i = 1; i <= m; ++i) {
    w(i - 1) = 0.5 * std::pow((Y.col(i - 1)).transpose() * (point), 2);
  }
  eigen_tail(w, point, n);
  w(m) = 1;

  Eigen::VectorXd Hw = Eigen::VectorXd::Zero(m + n + 1);
  Hw = Winv * w;

  int indexToBeReplaced = -1;
  double currentMax = -1;
  for (int j = 1; j <= m; ++j) {
    if (j == bestPointIndex) {
      continue;
    }
    double lagval = std::abs((Hw)(j - 1));
    if ((lagval > 1) || (norm(Y.col(j - 1) - bestPoint) > r * rho)) {
      double distance = norm((bestPoint - Y.col(j - 1)));
      double distanceWeight = std::pow(distance, 2);
      if (distance > 2 * rho) {
        distanceWeight += 100000000 * distanceWeight;
      }
      double value = distanceWeight * lagval;
      if (value >= currentMax) {
        indexToBeReplaced = j;
        currentMax = value;
      }
    }
  }
  return indexToBeReplaced;
}

double DFO_Model::norm(Eigen::VectorXd a) {
  if (normType == 0) {
    return (a).lpNorm<Infinity>();
  } else if (normType == 2) {
    return (a).norm();
  }
}

bool DFO_Model::FindReplacementForPointsOutsideRadius(double radius,
                                                      Eigen::MatrixXd &newPoints,
                                                      Eigen::VectorXi &newIndices) {
  Eigen::DiagonalMatrix<double, Eigen::Dynamic> copyS = S;
  Eigen::MatrixXd copyZ = Z;
  Eigen::MatrixXd copyUpsilon = Upsilon;
  Eigen::MatrixXd copyXi = Xi;
  Eigen::MatrixXd copyY = Y;
  Eigen::MatrixXd copyWinv = Winv;
  bool retVal = true;


  /// Find points outside r*radius
  Eigen::VectorXi sortedPoints = GetInterpolationPointsSortedByDistanceFromBestPoint();
  int number_of_points_outside = GetNumberOfPointsOutsideRadius(radius);
  sortedPoints.conservativeResize(number_of_points_outside);

  if (number_of_points_outside <= 0) {
    return false;
  }
  newIndices.resize(number_of_points_outside);

  for (int i = 0; i < number_of_points_outside; ++i) {
    newIndices(i) = -1;
  }


  subproblem.SetTrustRegionRadius((radius / r) * 0.9);
  newPoints.resize(n, number_of_points_outside);
  newPoints.setZero();
  Eigen::VectorXd dNew(n);
  int addedPoints = 0;
  int j = 0;
  for (int i = 0; i < number_of_points_outside; ++i) {
    dNew = FindLocalOptimumOfAbsoluteLagrangePolynomial(sortedPoints(i));
    double lagabsval = std::abs(ComputeLagrangePolynomial(sortedPoints(i), dNew));
    if (lagabsval > lagabsvalMin) {
      newIndices(addedPoints) = sortedPoints(i);
      newPoints.col(addedPoints) = dNew;
      eigen_col(newPoints, dNew, addedPoints);

      eigen_col(Y, dNew, sortedPoints(i) - 1);
      Winv = calculateWExplicitly();
      createW();
      addedPoints++;
      j++;
      //break;
    } else {
      PrintLagrangePolynomial(sortedPoints(i));
      if (newIndices.rows() == 0 || newIndices.rows() == addedPoints) {
        break;
      }
      //break;
    }
  }

  if (addedPoints != newIndices.rows()) {
    newIndices.conservativeResize(addedPoints);
    newPoints.conservativeResize(n, addedPoints);
  }
  if (addedPoints == 0) {
    retVal = false;
  }

  /// reset!!
  Y = copyY;
  Xi = copyXi;
  Upsilon = copyUpsilon;
  S = copyS;
  Z = copyZ;
  Winv = copyWinv;
  createW();
  return retVal;
}
int DFO_Model::GetNumberOfPointsOutsideRadius(double radius) {
  /// Number of points outside radius
  int number = 0;
  for (int j = 1; j <= m; ++j) {
    if (norm(Y.col(j - 1) - bestPoint) > radius) {
      number++;
    }
  }
  return number;
}


void DFO_Model::createW() {
  W = Eigen::MatrixXd::Zero(m + n + 1, m + n + 1);
  Eigen::MatrixXd A = Eigen::MatrixXd::Zero(m, m);
  Eigen::MatrixXd X = Eigen::MatrixXd::Zero(n + 1, m);
  for (int i = 1; i <= m; ++i) {
    for (int j = 1; j <= m; ++j) {
      A(i - 1, j - 1) = 0.5 * std::pow(Y.col(i - 1).transpose() * Y.col(j - 1), 2);
    }
  }

  for (int i = 1; i <= m; ++i) {
    X(0, i - 1) = 1;
  }

  for (int i = 1; i <= m; ++i) {
    Eigen::VectorXd tmp = X.col(i - 1);
    eigen_tail(tmp, Y.col(i - 1), n);
    eigen_col(X, tmp, i - 1);
  }

  W.topLeftCorner(m, m) = A;
  W.bottomLeftCorner(n + 1, m) = X;
  W.topRightCorner(m, n + 1) = X.transpose();
}

// This is slow and should only be used for testing and debugging.
Eigen::MatrixXd DFO_Model::calculateWExplicitly() {
  double precision = 0.00001;
  Eigen::MatrixXd W = Eigen::MatrixXd::Zero(m + n + 1, m + n + 1);
  Eigen::MatrixXd A = Eigen::MatrixXd::Zero(m, m);
  Eigen::MatrixXd X = Eigen::MatrixXd::Zero(n + 1, m);
  for (int i = 1; i <= m; ++i) {
    for (int j = 1; j <= m; ++j) {
      A(i - 1, j - 1) = 0.5 * std::pow(Y.col(i - 1).transpose() * Y.col(j - 1), 2);
    }
  }
  for (int i = 1; i <= m; ++i) {
    X(0, i - 1) = 1;
  }
  for (int i = 1; i <= m; ++i) {
    Eigen::VectorXd tmp = X.col(i - 1);
    eigen_tail(tmp, Y.col(i - 1), n);
    eigen_col(X, tmp, i - 1);
  }

  W.topLeftCorner(m, m) = A;
  W.bottomLeftCorner(n + 1, m) = X;
  W.topRightCorner(m, n + 1) = X.transpose();
  Eigen::MatrixXd Winv(m + n + 1, m + n + 1);
  Winv = W.inverse();
  Eigen::FullPivLU<Eigen::MatrixXd> lu_decompW(W);;
  Eigen::MatrixXd WLUinv(m + n + 1, m + n + 1);
  WLUinv = lu_decompW.inverse();
  return WLUinv;
}


void DFO_Model::UpdateOptimum() {
  int i = 1;
  for (int j = 2; j <= m; ++j) {
    if (fvals[j - 1] < fvals[i - 1]) {
      i = j;
    }
  }
  if (i != bestPointIndex) {
    bestPointIndex = i;
    bestPoint = Y.col(i - 1);
  }
}

bool DFO_Model::isPoised(Eigen::VectorXd &dNew, int &indexOfPointToBeReplaced, double radius) {
  bool ispoised = false;
  indexOfPointToBeReplaced = -1;
  int numberOfPointsOutsideRadius = GetNumberOfPointsOutsideRadius(radius);
  if (numberOfPointsOutsideRadius >= 1) {
    // Find points outside r*radius
    Eigen::VectorXi sortedPoints = GetInterpolationPointsSortedByDistanceFromBestPoint();
    sortedPoints.conservativeResize(numberOfPointsOutsideRadius);
    subproblem.SetTrustRegionRadius(radius / r);
    for (int i = 0; i < numberOfPointsOutsideRadius; ++i) {
      dNew = FindLocalOptimumOfAbsoluteLagrangePolynomial(sortedPoints[i]);
      double lagabsval = std::abs(ComputeLagrangePolynomial(sortedPoints[i], dNew));
      if (lagabsval > lagabsvalMin) {
        indexOfPointToBeReplaced = sortedPoints[i];
        break;
      } else {

        PrintLagrangePolynomial(sortedPoints[i]);
      }
    }
  } else {
    subproblem.SetTrustRegionRadius(radius);
    findWorstPointInInterpolationSet(dNew, indexOfPointToBeReplaced);
  }

  if (indexOfPointToBeReplaced == -1) {
    isModelCFL = 1;
    ispoised = true;
  } else {
    isModelCFL = 0;
  }
}

void DFO_Model::modelImprovementStep(Eigen::VectorXd &dNew, int &indexOfPointToBeReplaced) {
  indexOfPointToBeReplaced = -1;
  Eigen::VectorXi sortedPoints = GetInterpolationPointsSortedByDistanceFromBestPoint();
  for (int i = 0; i < m; ++i) {
    int t = sortedPoints[i];
    subproblem.SetTrustRegionRadius(rho);
    dNew = FindLocalOptimumOfAbsoluteLagrangePolynomial(t);
    double lagabsval = std::abs(ComputeLagrangePolynomial(t, dNew));
    if (lagabsval > lambda || norm((Y.col(t - 1) - bestPoint))
        > r * rho) {

      if (lagabsval > lagabsvalMin) {
        indexOfPointToBeReplaced = t;
        break;
      }
    }
  }

  if (indexOfPointToBeReplaced == -1) {
    isModelCFL = 1;
  } else {
    isModelCFL = -1;
  }
}

void DFO_Model::isLagrangePoly() {
  for (int t = 1; t <= m; ++t) {

    double c;
    Eigen::VectorXd grad(n);
    Eigen::MatrixXd hess(n, n);
    createLagrangePolynomial(t, c, grad, hess);

    for (int j = 1; j <= m; j++) {
      double val = c + grad.transpose() * Y.col(j - 1) + 0.5 * (Y.col(j - 1)).transpose() * hess * Y.col(j - 1);
      if (j == t) {
        std::cout << "value should be 1, but is: " << val << "\n";
      } else {
        std::cout << "value should be 0, but is: " << val << "\n";
      }
    }
  }
}

void DFO_Model::Converged(int iterations,
                          int number_of_tiny_improvements,
                          int number_of_function_calls,
                          int number_of_parallell_function_calls) {
  {
    Eigen::VectorXd gradient = GetLagrangianGradient(GetBestPoint());
    Eigen::MatrixXd Yabs(n, m);
    for (int j = 0; j < m; ++j) {
      Eigen::VectorXd sd = (Y).col(j) + getCenterPoint();
      eigen_col(Yabs, sd, j);
    }
    std::cout.clear();
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;
    std::cout << "\033[1;36;mDFO terminated. Trust region radius too small.\033[0m" << std::endl;

    std::cout << "\033[1;34;m " << "Norm of gradient at best point = " << "\033[0m" << gradient.norm() << "\n";
    std::cout << "\033[1;34;m " << "Best point index = " << "\033[0m" << bestPointIndex << "\n";
    std::cout << "\033[1;34;m " << "Fvals = \n" << "\033[0m" << fvals << "\n";
    std::cout << "\033[1;34;m " << "Y = \n" << "\033[0m" << Y << "\n";
    std::cout << "\033[1;34;m " << "Y absolute = \n" << "\033[0m" << Yabs << "\n";
    std::cout << "\033[1;34;m " << "Ybest (abs) = \n" << "\033[0m" << bestPoint + y0 << "\n";
    std::cout << "\033[1;34;m " << "Trust region radius is: " << "\033[0m" << rho << std::endl;;
    std::cout << "\033[1;34;m " << "Best found, all time, value: " << "\033[0m" << bestPointAllTimeFunctionValue
              << "\n";
    std::cout << "\033[1;34;m " << "Best found, all time, point: \n" << "\033[0m" << bestPointAllTime + y0;

    std::cout << "\nm = " << m << "\n";
    std::cout << "n = " << n << "\n";
    std::cout << "ng = " << ng << "\n";

    std::cout << "Best point (absolute):\n" << getCenterPoint() + GetBestPoint()
              << "\nWith value: " << GetFunctionValue(getBestPointIndex()) << "\n";

    std::cout << "\033[1;36;mFunction Calls \033[0m" << number_of_function_calls << "\n";
    std::cout << "\033[1;36;mParallell Function Calls \033[0m" << number_of_parallell_function_calls << "\n";
  }
}

void DFO_Model::createLagrangePolynomial(int t, double &c, Eigen::VectorXd &grad, Eigen::MatrixXd &hess) {
  grad.setZero();
  hess.setZero();
  c = 0;
  createW();
  Eigen::VectorXd ans(n + m + 1);
  Eigen::VectorXd rhs = Eigen::VectorXd::Zero(n + m + 1);
  rhs(t - 1) = 1;
  ans = W.colPivHouseholderQr().solve(rhs);

  c = ans(m);
  grad = ans.tail(n);
  for (int k = 1; k <= m; ++k) {
    hess += ans(k - 1) * (Y.col(k - 1)) * (Y.col(k - 1)).transpose();
  }
}

// return true when no new points are found. is_poised = true
bool DFO_Model::ModelImprovementAlgorithm(double radius, Eigen::MatrixXd &newPoints, Eigen::VectorXi &newIndices) {
  /// Get all points inside of the trust-region
  Eigen::MatrixXd tmpNewPoints;
  Eigen::VectorXi tmpNewIndices;
  FindReplacementForPointsOutsideRadius(radius, tmpNewPoints, tmpNewIndices);
  Eigen::MatrixXd copyY = Y;
  Eigen::VectorXi changed(m);
  changed.setZero();
  for (int i = 0; i < tmpNewPoints.cols(); ++i) {
    if (tmpNewIndices(i) != -1) {
      eigen_col(Y, tmpNewPoints.col(i), tmpNewIndices(i) - 1);
      changed(tmpNewIndices(i) - 1) = 1;
    } else {
      break;
    }
  }
  subproblem.SetTrustRegionRadius(radius);
  /// Improve poisedness until required poisedness is achieved.
  Eigen::VectorXd dNew(n);
  int index = -1;
  while (1) {
    findWorstPointInInterpolationSet(dNew, index);
    if (index == -1) {
      break;
    } else {
      eigen_col(Y, dNew, index - 1);
      changed(index - 1) = 1;
    }
  }

  int number_of_new_points = 0;
  for (int i = 0; i < m; i++) {
    if (changed(i) == 1) {
      number_of_new_points++;
    }
  }
  newPoints.resize(n, number_of_new_points);
  newIndices.resize(number_of_new_points);

  int i = 0;
  for (int j = 1; j <= m; ++j) {
    if (changed(j - 1) == 1) {
      eigen_col(newPoints, Y.col(j - 1), i);
      newIndices(i) = j;
      i++;
    }
  }
  Y = copyY; /// reset.
  createW();

  if (number_of_new_points == 0) {
    isModelCFL = 1;
  } else {
    isModelCFL = 0;
  }

  return (number_of_new_points == 0);
}

// Will only be valid for bestpoint, because the lagrange multipliers are found for that point.
Eigen::VectorXd DFO_Model::GetLagrangianGradient(Eigen::VectorXd point) {
  Eigen::VectorXd lagGrad(n);
  lagGrad = GetGradientAtPoint(point);
  subproblem.SetTrustRegionRadius(rho);
  if (ng > 0) {
    calculateLagrangeMultipliers();
    Eigen::MatrixXd conGrad = subproblem.getGradientConstraints(point + y0);
    for (int i = 0; i < subproblem.getNumberOfConstraints(); i++) {
      lagGrad -= lagrangeMultipliers[i] * ((conGrad.row(i)).transpose());
    }
  }
  return lagGrad;
}

void DFO_Model::calculateLagrangeMultipliers() {

  subproblem.calculateLagrangeMultipliers((char *) "Minimize", y0, bestPoint, bestPoint);
  lagrangeMultipliers = subproblem.getLagrangeMultipliers();

}

Eigen::VectorXd DFO_Model::ScaleVariablesFromApplicationToAlgorithm(Eigen::VectorXd point) {
  Eigen::VectorXd ret(point.rows());
  point = point / 10000.0;
  for (int i = 0; i < point.rows(); ++i) {
    if (i % 2 == 0 && i != 0) {
      //z coord
      ret[i] = point[i] * 30.0;
    } else {
      //xy coord
      ret[i] = point[i];
    }
  }
  return ret;

}
void DFO_Model::setInitialy0(Eigen::VectorXd a) {

  this->y0 = a;
}
void DFO_Model::findVariableMeaning(Eigen::VectorXd realvars, Eigen::VectorXd scaling) {
  mapNormalToFieldopt = Eigen::VectorXi::Zero(6); // [x0 y0 z0, x1 y1 z1]
  for (int i = 0; i < 6; i++) {
    for (int j = 0; j < 6; j++) {
      if (abs(copyOfStartingPoint[i] - realvars[j]) <= 0.000001) {
        mapNormalToFieldopt[i] = j;
      }
    }
  }
  subproblem.SetMappingVariables(mapNormalToFieldopt, scaling);
}

void DFO_Model::PrintSortedBestPoint(Eigen::VectorXd scaling) {
  Eigen::VectorXd wellStart(3);
  Eigen::VectorXd wellEnd(3);
  Eigen::VectorXd splineIsh = y0 + bestPoint;

  for (int i = 0; i < 6; ++i) {
    splineIsh[i] = splineIsh[i] * scaling[i];
  }
  for (int i = 0; i < 3; i++) {
    wellStart[i] = splineIsh(mapNormalToFieldopt[i]);
    wellEnd[i] = splineIsh(mapNormalToFieldopt[i + 3]);
  }
  std::cout << "Well information\nCoordinates:\n" << wellStart << "\n" << wellEnd << "\n" << "spline length: "
            << (wellStart - wellEnd).norm() << "\n\n";
}

}
}