
#ifndef FIELDOPT_DFO_MODEL_H
#define FIELDOPT_DFO_MODEL_H

#include <iostream>
#include <Eigen/Dense>
#include <random>
#include <math.h>
#include <Settings/optimizer.h>
#include <Subproblem.h>
#include "Subproblem.h"
#include "EigenUtil.h"
#include "GradientEnhancedModel.h"
/* References
This implementation is based upon the following papers and book, and I would recommend anyone who
is trying to understand the code to actively use those references.
[1] The BOBYQA algorithm for bound constrained optimization without derivatives by M.J.D. Powell.
[2] The NEWUOA software for unconstrained optimization without derivatives by M.J.D. Powell.
[3] Introduction to Derivative-Free Optimization by Andrew R. Conn, Katya Scheinberg and Luis N. Vicente.
[4] Least Frobenius norm updating of quadratic models that satisfy interpolation conditions by M.J.D. Powell.

Only the parts about model improvement (i.e., finding upper/lower/high value of the Lagrange polynomials,
are based upon [3]. Almost everything else is based upon [2] (practical approach) and [4] (more theoretical view).
*/
namespace Optimization {
namespace Optimizers {

class DFO_Model {

 private:

  Eigen::VectorXd copyOfStartingPoint;
  Eigen::VectorXi mapNormalToFieldopt;

  Eigen::VectorXd lagrangeMultipliers;
  const int normType;
  double lagabsvalMin = 0.5; // works ok: 0.001

  Settings::Optimizer *settings_;
  GradientEnhancedModel enhancedModel;
  Subproblem subproblem;
  unsigned int m; // Number of interpolation points used to create the model. Does not change.
  unsigned int n; // Number of decision variables in your model.
  unsigned int ng; // Number of decision variables WITH gradients.
  double rho; // Trust-region radius.
  double lambda; // The required poisedness of the set of interpolation points.
  double r;

  Eigen::MatrixXd Winv;
  Eigen::MatrixXd W;

  Eigen::VectorXd y0; // The point which the model is ceneterd around.

  Eigen::MatrixXd
      Y;  // Container for the interpolation points. The interpolation point i is given in the following way:
  // yi = y0 + Y.col(i); i.e. Y contains the displacements away from y0.

  Eigen::MatrixXd derivatives;
  Eigen::VectorXd derivativeAtCenterpoint;

  Eigen::VectorXd fvals; // Holds the function evaluations for the interpolation points.

  Eigen::VectorXd bestPoint; // Displacement of the optimal point from y0. Absolute point: y0 + bestPoint;
  int
      bestPointIndex; // This assumes that the point is in the interpolation set (not sure if this will be discarded in future work)

  Eigen::VectorXd bestPointAllTime;
  double bestPointAllTimeFunctionValue;

  //These 4 variables are only used in the case m>2*n+1. How to calculate the p's and q's are well explained in [1], for sigma see [2].
  //The i's are used to find the corresponding interpolation points and function values in Y and Fvals, respectively.
  //The qs, qs and is come in a set of 3, such that the same index of each vector belong together.
  Eigen::VectorXi sigmas;
  Eigen::VectorXi ps;
  Eigen::VectorXi qs;
  Eigen::VectorXi is;

  // The inverse KKT matrix, H. See [2]
  Eigen::MatrixXd Xi;
  Eigen::MatrixXd Upsilon;
  Eigen::MatrixXd Z;
  Eigen::DiagonalMatrix<double, Eigen::Dynamic>
      S; //Should have been a integer matrix, but Eigen doesn't support diagonal int matrices.


  // Containers for the 2nd order model.
  Eigen::VectorXd gradient;
  Eigen::VectorXd centerPoint;
  Eigen::MatrixXd hessian;
  double constant = 0;

  Eigen::MatrixXd Gamma;
  Eigen::VectorXd gammas;

  bool modelInitialized;
  bool initialInterpolationPointsFound;

  int isModelCFL = -1; // -1: Don't know.     0: No.    1: Yes.

  /**
  Checks if a value is almost zero.

  @value the value to be checked.
  @zeroLimit the highest value that still is counted as zero.
  @return true / false
  */
  bool isApproxZero(double value, double zeroLimit);

  /**
  The classic kronecker-delta function.

  Compares if two integers are equal or not.

  @param[in] i integer 1
  @param[in] j integer 2
  @return 1 if i==j and zero otherwise.
  */
  int kroneckerDelta(int i, int j);

  /**
  Returns the sign of the value.

  0 is counted as positive (+1).

  @param[in] the value to be checked.
  @return the sign of the value.
  */
  int sign(double value);



  bool cmp(Eigen::VectorXd a, Eigen::VectorXd b);

 public:
  //EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  enum UpdateReason {
    IMPROVE_POISEDNESS = -1,
    INCLUDE_NEW_OPTIMUM = -2,
    INCLUDE_NEW_POINT = -3,
    FORCED_IMPROVE_MODEL = -4
  };

  bool isInitialInterpolationPointsFound() {
    return initialInterpolationPointsFound;
  }

  bool isModelInitialized() {
    return modelInitialized;
  }

  /**
  Constructor for the class.

  A naming convenvention:
  model = this entire class.
  quadratic model = just the quadratic model.
  Inverse KKT matrix = H

  @param[in] m number of interpolation points.
  @param[in] n number of decision variables.
  @param[in] y0 the center point of the model.
  @param[in] rhoBeg the initial trust-region radius.
  @param[in] lambda the required poisedness of the interpolation set (lambda > 1)
  */
  DFO_Model(unsigned int m,
            unsigned int n,
            unsigned int ng,
            Eigen::VectorXd y0,
            double rhoBeg,
            double lambda,
            double weight_objective_minimum_change,
            QList<double> weights_derivatives,
            Settings::Optimizer *settings);

  DFO_Model() : normType(0) {};

  Eigen::VectorXd ScaleVariablesFromApplicationToAlgorithm(Eigen::VectorXd point);

  /**
  Finds the first set of interpolation points.

  Finds the first set of interpolation points based
  upon the initial center point (y0) and the trust-region radius.

  */
  Eigen::MatrixXd findFirstSetOfInterpolationPoints();

  /**
  Finds the last set of interpolation points.

  If m > 2n+1, then this function must also be ran. Before it is ran,
  make sure that the local variable fvals is filled up with the function
  evaluations for the first set of interpolation points.

  */
  Eigen::MatrixXd findLastSetOfInterpolationPoints();

  /**
  Initializes the model.

  Both findFirstSetOfInterpolationPoints() and findLastSetOfInterpolationPoints() (if m > 2n+1)
  must be called before this function.
  */
  void initializeModel();

  /**
  Updates the model with a new point.

  Updates the model with a new point, the reason for updating must be provided.
  Some changes must be done, and this function is NOT done!
  Cannot be used when INCLUDE_NEW_OPTIMUM is the reason.

  @param[in] yNew is the  is the displacement of the new point from the current center point(y0).
  @param[in] fvalNew is the function evaluation corresponding to yNew.
  @param[in] t-1 is the index of the point that is going to be replaced by yNew in Y.
  @param[in] updateReason is either IMPROVE_POISEDNESS or INCLUDE_NEW_OPTIMUM.
  */
  void update(Eigen::VectorXd yNew,
              double fvalNew,
              Eigen::VectorXd gradient,
              unsigned int t,
              UpdateReason updateReason);

  void update(Eigen::MatrixXd yNews,
              Eigen::VectorXd fvalNews,
              Eigen::MatrixXd gradients,
              Eigen::VectorXi indicies,
              int numberOfPoints,
              UpdateReason updateReason);

  /**
  Evaluates the current quadratic model at point.

  @param[in] point is the displacement from current center point (y0).
  @return the value of the model at point.
  */
  double evaluateQuadraticModel(Eigen::VectorXd point);

  /**
  This is a bad way of accessing and updating the function evaluations.

  It will be replaced by other functions later, when I have decided how
  the interface should be.
  @return a reference to the fvals
  */
  Eigen::VectorXd *getFvalsReference() {
    return &(this->fvals);
  }

  /**
  This is a bad way of accessing Y outside of the class.

  It will be replaced by other functions later, when I have decided how
  the interface should be.
  @return a reference to the Y
  */
  Eigen::MatrixXd *getYReference() {
    return &(this->Y);
  }

  Eigen::MatrixXd *getDerivativeReference() {
    return &(this->derivatives);
  }

  /**
  Returns the center point of the quadratic model

  @return the center point of the quadratic model
  */
  Eigen::VectorXd getCenterPoint() {
    return y0;
  }


  void findWorstPointInInterpolationSet(Eigen::VectorXd &dNew, int &indexOfWorstPoint);
  void calculateAMatrix(Eigen::MatrixXd &A, Eigen::MatrixXd &Ycopy);


  Eigen::VectorXd FindLocalOptimumOfAbsoluteLagrangePolynomial(int t);

  /**
  Finds the point that is best to replace with the new one.

  @yNew is the point that we want to add to the model, where yNew is given as the displacement of the current center point.
  @return index of the point that is best to replace.
  */
  int findPointToReplaceWithNewOptimum(Eigen::VectorXd yNew);

  /**
  Returns the index of the best point.

  Note: this will only be valid as long as the best yet found
  point is never removed from the interpolation set.

  @return the index of the best point.
  */
  int getBestPointIndex();

  void SetFunctionValue(int t, double value);
  void SetFunctionValueAndDerivatives(int t, double value, Eigen::VectorXd grad);
  void SetTrustRegionRadiusForSubproblem(double radius);
  bool FindReplacementForPointsOutsideRadius(double radius, Eigen::MatrixXd &newPoints, Eigen::VectorXi &newIndices);
  double GetFunctionValue(int t) {
    return fvals[t - 1];
  }

  double GetTrustRegionRadius() {
    return rho;
  }

  void SetTrustRegionRadius(double radius) {
    rho = radius;
  }

  void SetRequiredPoisedness(double lambda) {
    this->lambda = lambda;
  }
  void SetInitialStartPoint(Eigen::VectorXd startPoint) {
    y0 = startPoint;
  }
  void SetNumberOfVariables(int n) {
    this->n = n;
  }
  void SetNumberOfInterpolationPoints(int m) {
    this->m = m;
  }

  Eigen::VectorXd GetGradient() {
    return gradient;
  }
  Eigen::VectorXd GetGradientAtPoint(Eigen::VectorXd point) {
    enhancedModel.ComputeModel(Y, derivatives, derivativeAtCenterpoint, fvals, y0, bestPoint, rho, r, ng);
    double e_c = 0;
    Eigen::VectorXd e_g(n);
    Eigen::MatrixXd e_h(n, n);
    enhancedModel.GetModel(e_c, e_g, e_h);
    return e_g + e_h * point;

  }
  Eigen::VectorXd GetPoint(int t) {
    return Y.col(t - 1);
  }
  Eigen::VectorXd GetBestPoint() {
    return Y.col(bestPointIndex - 1);
  }
  double GetBestFunctionValueAllTime() {
    return bestPointAllTimeFunctionValue;
  }
  Eigen::VectorXd GetBestPointAllTime() {
    return bestPointAllTime;
  }
  Eigen::VectorXd FindLocalOptimum();
  double findLargestDistanceBetweenPointsAndOptimum();
  double ComputeLagrangePolynomial(int t, Eigen::VectorXd point);
  double PrintLagrangePolynomial(int t);
  Eigen::VectorXi GetInterpolationPointsSortedByDistanceFromBestPoint();
  bool FindPointToReplaceWithPointOutsideScaledTrustRegion(int t, Eigen::VectorXd &dNew);
  void wtf(Eigen::VectorXd &da) {
    return;
  }
  int isPointAcceptable(Eigen::VectorXd point);
  int GetNumberOfPointsOutsideRadius(double radius);
  double norm(Eigen::VectorXd a);
  Eigen::MatrixXd calculateWExplicitly();
  void UpdateOptimum();
  bool isPoised(VectorXd &dNew, int &indexOfPointToBeReplaced, double radius);
  void modelImprovementStep(VectorXd &dNew, int &indexOfPointToBeReplaced);
  void Converged(int iterations,
                 int number_of_tiny_improvements,
                 int number_of_function_calls,
                 int number_of_parallell_function_calls);
  void isLagrangePoly(); // prints the (hopefully) kronecker-delta property. used for debugging.
  void createLagrangePolynomial(int t, double &c, VectorXd &grad, MatrixXd &hess);
  void updateQuadraticModelNew(Eigen::VectorXd yNew, double fvalNew, unsigned int t);
  void shiftCenterPointOfQuadraticModelNew(Eigen::VectorXd s);
  void createW();
  int IsModelCFL() {
    return isModelCFL;
  }
  Eigen::VectorXd GetLagrangianGradient(Eigen::VectorXd point);
  void initLagrangeMultipliers(int a) {
    lagrangeMultipliers = Eigen::VectorXd::Zero(a);
  }
  bool ModelImprovementAlgorithm(double radius, Eigen::MatrixXd &newPoints, Eigen::VectorXi &newIndices);
  void calculateLagrangeMultipliers();
  //These two functions are only used to create an example to show how the model improvement algorithm works
  void setInitialy0(Eigen::VectorXd);
  void findVariableMeaning(Eigen::VectorXd realvars, Eigen::VectorXd scaling);
  void PrintSortedBestPoint(Eigen::VectorXd scaling);

};
}
}

#endif //FIELDOPT_DFO_MODEL_H