//
// Created by joakim on 24.04.18.
//

#include <iostream>
#include "GradientEnhancedModel.h"
#include "EigenUtil.h"

namespace Optimization {
namespace Optimizers {

static Eigen::VectorXd _y0_;
static Eigen::MatrixXd _hessian_old_;
static Eigen::MatrixXd _D_;
static Eigen::MatrixXd _v_;
static double _alpha_;
static Eigen::VectorXd _weights_least_square_;
static Eigen::MatrixXd _points_;
static Eigen::VectorXd _best_point_;
static int _n_;
static int _ng_;
static int _m_;

void GradientEnhancedModel::GetConstant(double &c) {
  c = constant_;
}
void GradientEnhancedModel::GetGradient(Eigen::VectorXd &g) {
  g = gradient_;
}
void GradientEnhancedModel::GetHessian(Eigen::MatrixXd &H) {
  H = hessian_;
}
void GradientEnhancedModel::GetModel(double &c, Eigen::VectorXd &g, Eigen::MatrixXd &H) {
  c = constant_;
  g = gradient_;
  H = hessian_;
}
GradientEnhancedModel::GradientEnhancedModel(
    int n, int m, int number_of_variables_with_gradient,
    QList<double> weights_derivatives,
    double weight_objective_minimum_change) {

  _n_ = n;
  n_ = n;
  _m_ = m;
  m_ = m;
  _ng_ = number_of_variables_with_gradient;
  ng_ = number_of_variables_with_gradient;
  alpha_ = weight_objective_minimum_change;
  weights_derivatives_ = Eigen::VectorXd::Zero(weights_derivatives.size());
  int j = 0;
  for (auto i = weights_derivatives.begin(); i != weights_derivatives.end(); ++i) {
    weights_derivatives_[j] = *i;
    j++;
  }
  constant_ = 0;
  gradient_ = Eigen::VectorXd::Zero(n_);
  hessian_ = Eigen::MatrixXd::Zero(n_, n_);
  hessian_old_ = Eigen::MatrixXd::Zero(n_, n_);
  points_ = Eigen::MatrixXd::Zero(n_, m_);
  funcVals_ = Eigen::VectorXd::Zero(n_);
  v_ = Eigen::VectorXd::Zero(ng_ * m_);
  y0_ = Eigen::VectorXd::Zero(n_);
  best_point_ = Eigen::VectorXd::Zero(n_);
  weights_least_square_ = Eigen::VectorXd::Zero(m_);

  int t = 0;
  int y = n_;
  for (int i = 0; i < ng_; ++i) {
    t += y;
    y--;
  }
  h_old_ = Eigen::VectorXd::Zero(t);
  D_ = Eigen::MatrixXd::Zero((ng_) * m_, t);

}

void GradientEnhancedModel::ComputeModel(Eigen::MatrixXd Y,
                                         Eigen::MatrixXd derivatives,
                                         Eigen::VectorXd derivatives_at_y0,
                                         Eigen::VectorXd funcVals,
                                         Eigen::VectorXd y0,
                                         Eigen::VectorXd best_point,
                                         double radius, double scaling_factor_r,
                                         int index_of_center_point) {
  funcVals_ = funcVals;
  points_ = Y;

  y0_ = y0;
  best_point_ = best_point;

  // Set up _weights_least_square_
  for (int t = 0; t < m_; ++t) {
    double norm = (Y.col(t) - best_point).norm();
    if (norm <= radius) {
      weights_least_square_[t] = weights_derivatives_[0];
    } else if (norm <= scaling_factor_r * radius) {
      weights_least_square_[t] = weights_derivatives_[1];
    } else {
      weights_least_square_[t] = weights_derivatives_[2];
    }
  }

  // Set the _D_ matrix and the _v_ vector
  if (ng_ > 0) {
    int base_row = 0;
    for (int t = 0; t < m_; ++t) { // for each sample point

      int y = n_;
      int c0 = 0;
      for (int i = 0; i < ng_; ++i) { // for each row
        for (int k = 0; k < y; ++k) { // for each elem in row
          D_(base_row + i, c0 + k) = points_(k, t) * weights_least_square_[t];
        }
        c0 += y;
        y--;
      }
      y = n_ - 1;
      c0 = 0;
      for (int k = 0; k < (ng_ - 1); ++k) { // for each row that begins an inverse diagonal
        int ii = 1;
        int jj = 1;
        for (int i = k; i < (ng_ - 1); ++i) {//for each element in that inverse diagonal
          D_(base_row + k + ii, c0 + y - jj) = points_(y, t) * weights_least_square_[t];
          ii++;
          jj++;
        }
        c0 += (y + 1);
        y--;
      }
      base_row += ng_;

      for (int i = 0; i < ng_; ++i) {
        v_(t * ng_ + i) = (derivatives((ng_ - i - 1), t) - derivatives_at_y0(ng_ - i - 1)) * weights_least_square_[t];
      }
    }
  }


  // set the constraints
  Eigen::VectorXd ans;
  solveLinearSystem(funcVals, derivatives_at_y0, ans);

  // calculate start indices;
  int start_h_ij = 0;
  int start_g_i = (int) ((n_ * n_ + n_) * 0.5);
  int start_c = start_g_i + (n_ - ng_);
  int start_lambda_i = start_c + 1;

  //extract the results
  constant_ = ans(start_c);
  for (int i = 0; i < n_ - ng_; ++i) {
    gradient_[i] = ans(start_g_i + i);
  }
  eigen_tail(gradient_, derivatives_at_y0, ng_);
  for (int i = 1; i <= (int) ((n_ * n_ + n_) * 0.5); ++i) {
    int ii = 0;
    int jj = 0;
    convert_t_to_ij_vectorized(i, ii, jj);

    hessian_(ii - 1, jj - 1) = ans(start_h_ij + i - 1);
    if (ii != jj) {
      hessian_(jj - 1, ii - 1) = hessian_(ii - 1, jj - 1);
    }
  }

  hessian_old_ = hessian_;
}

void GradientEnhancedModel::solveLinearSystem(Eigen::VectorXd funcVals,
                                              Eigen::VectorXd derivatives_at_y0,
                                              Eigen::VectorXd &ans) {
  int colsD = 0;
  int y = n_;
  for (int i = 0; i < ng_; ++i) {
    colsD += y;
    y--;
  }
  Eigen::MatrixXd A = Eigen::MatrixXd::Zero((int) (1 + (n_ - ng_) + m_ + (n_ * n_ + n_) * 0.5),
                                            (int) (1 + (n_ - ng_) + m_ + (n_ * n_ + n_) * 0.5));
  Eigen::VectorXd b = Eigen::VectorXd::Zero((int) (1 + (n_ - ng_) + m_ + (n_ * n_ + n_) * 0.5));
  // calculate start indices;
  int start_h_ij = 0;
  int start_g_i = (int) ((n_ * n_ + n_) * 0.5);
  int start_c = start_g_i + (n_ - ng_);
  int start_lambda_i = start_c + 1;
  int row = 0;




  // dL/dhij,
  for (int i = 1; i <= n_; ++i) {
    for (int j = 1; j <= i; ++j) { // the derivative with respect to (i,j)

      b(row) += hessian_old_(i - 1, j - 1) * alpha_; //right hand side
      A(row, convert_h_ij_to_t_vectorized(i, j) - 1) = alpha_;

      for (int t = 1; t <= m_; ++t) {
        A(row, start_lambda_i + t - 1) += -0.5 * points_(i - 1, t - 1) * points_(j - 1, t - 1);
      }

      if (i > (n_ - ng_)) {
        for (int p = 1; p <= ng_ * m_; p++) {
          int t = convert_h_ij_to_t_lsq(i, j); // taking derivative w.r.t. h_t

          b(row) += (1 - alpha_) * v_(p - 1) * D_(p - 1, t - 1); // right hand side

          for (int k = 1; k <= colsD; ++k) {
            int ii = 0;
            int jj = 0;

            convert_t_to_ij_lsq(k, ii, jj);
            A(row, convert_h_ij_to_t_vectorized(ii, jj) - 1) += (1 - alpha_) * D_(p - 1, k - 1) * D_(p - 1, t - 1);
          }
        }

      }

      row++;
    }
  }
  // dL/dgi,
  for (int i = 1; i <= (n_ - ng_); ++i) {
    for (int t = 1; t <= m_; ++t) {
      A(row, start_lambda_i + t - 1) += points_(i - 1, t - 1);
    }
    row++;
  }
  // dL/dc
  for (int t = 1; t <= m_; ++t) {
    A(row, start_lambda_i + t - 1) = 1;
  }
  row++;

  //interpolation conditions
  for (int k = 1; k <= m_; ++k) {
    b(row) = funcVals(k - 1) - derivatives_at_y0.dot((points_.col(k - 1)).tail(ng_));
    A(row, start_c) = 1;

    for (int t = 1; t <= (n_ - ng_); ++t) {
      A(row, start_g_i + t - 1) = points_(t - 1, k - 1);
    }

    for (int i = 1; i <= n_; ++i) {
      for (int j = 1; j <= i; ++j) {
        if (i == j) {
          A(row, start_h_ij + convert_h_ij_to_t_vectorized(i, j) - 1) =
              0.5 * points_(i - 1, k - 1) * points_(i - 1, k - 1);
        } else {
          A(row, start_h_ij + convert_h_ij_to_t_vectorized(i, j) - 1) = points_(i - 1, k - 1) * points_(j - 1, k - 1);
        }
      }
    }

    row++;
  }
  ans = A.colPivHouseholderQr().solve(b);
}

int GradientEnhancedModel::convert_h_ij_to_t_lsq(int i, int j) {
  int t = 0;
  for (int p = n_; p > i; --p) {
    t += p;
  }
  t += j;
  return t;
}

int GradientEnhancedModel::convert_h_ij_to_t_vectorized(int i, int j) {
  int t = 0;
  int y = n_;
  for (int k = 1; k < j; ++k) {
    t += y;
    y--;
  }
  t += (i - j + 1);
  return t;
}

void GradientEnhancedModel::convert_t_to_ij_lsq(int t, int &i, int &j) {
  i = n_;
  j = 0;
  for (i = n_; t - i > 0; --i) {
    t -= i;
  }
  j = t;
}

void GradientEnhancedModel::convert_t_to_ij_vectorized(int t, int &i, int &j) {
  j = 1;
  for (int k = n_; t > k; k--) {
    t = t - k;
    j++;
  }
  i = j + (t - 1);
}

void GradientEnhancedModel::isInterpolating() {
  std::cout << "Checking if the gradient enhanced model actually interpolates the points\n";
  for (int i = 1; i <= m_; ++i) {
    double val = constant_ + gradient_.transpose() * points_.col(i - 1)
        + 0.5 * (points_.col(i - 1)).transpose() * hessian_old_ * points_.col(i - 1);
    std::cout << funcVals_[i - 1] << " == " << val << "\t" << (std::abs(val - funcVals_[i - 1]) <= 0.000000001) << "\n";
  }
}

}
}